#! python

# imports:

import numpy as np
import matplotlib.pyplot as plt
from scipy import interpolate


def XMCD(
    pdat,
    mdat,
    ene,
    det,
    mon,
    log=False,
    xmin=None,
    xmax=None,
    xsize=10000,
    norm="edge_jump",
    eshift=0,
) -> tuple:
    """
    xray magnetic circular dichroism function

    Arguments
    ---------

        pdat    = positive list of pandas objects
        mdat    = negative list of pandas objects
        ene     = energy column name of pandas object
        det     = detector column name of pandas object
        mon     = monitor column -> if no monitor wanted, put 'False'
        log     = boolean, use logarithm for transmission experiments
        xmin    = adjust minimum energy
        xmax    = adjust maximum energy
        xsize   = number of points for interpolation default 10000
                  ! be careful: low point density can lead to ValueErrors due
                  ! to rounding numbers,
                  ! set to None, if no interpolation is needed/wanted
        norm    = choose normalization from ['white_line', 'edge_jump',
                  'pre_edge','None']
        eshift  = shift in energy, added to pdat values, default 0

    Returns
    --------

    energy, plus_xas, minus_xas, averaged_xas, xmcd

    Notes
    --------

    no notes
    """

    t2pdat = pdat
    t2mdat = mdat
    # photon energies:
    pmin = []
    pmax = []
    for n in range(len(t2pdat)):
        pmin.append(np.min(t2pdat[n][ene]))
        pmax.append(np.max(t2pdat[n][ene]))
    mmin = []
    mmax = []
    for n in range(len(t2mdat)):
        mmin.append(np.min(t2mdat[n][ene]))
        mmax.append(np.max(t2mdat[n][ene]))

    # used energy range:
    if xmin is None:
        xmin = np.max(pmin + mmin)
    if xmax is None:
        xmax = np.min(pmax + mmax)

    # energy interpolation range:
    if xsize is None:
        xx = np.array(t2pdat[n][ene])
    else:
        xx = np.linspace(xmin + 0.1, xmax - 0.1, xsize)
    t3pdat = []  # list to addinterpolated
    for n in range(len(t2pdat)):
        v1 = np.array(t2pdat[n][ene])
        if mon is not False:
            v21 = np.array(t2pdat[n][det])
            v22 = np.array(t2pdat[n][mon])
            if log is True:
                v2 = np.log(v22 / v21)
            else:
                v2 = v21 / v22
        if mon is False:
            v2 = np.array(t2pdat[n][det])
        if xsize is None:
            t3pdat.append(v2)
        else:
            try:
                t3pdat.append(
                    interpolate.interp1d(v1, v2)(xx + float(eshift) / 1000)
                )
            except Exception:
                raise ValueError(
                    'error at plus helicity; check interpolation range ' +
                    'eshift to large for interpolation range? ' +
                    'max Eshift ca. 100meV.'
                )
    t3mdat = []
    for n in range(len(t2mdat)):
        v1 = np.array(t2mdat[n][ene])
        if mon is not False:
            v21 = np.array(t2mdat[n][det])
            v22 = np.array(t2mdat[n][mon])
            if log is True:
                v2 = np.log(v22 / v21)
            else:
                v2 = v21 / v22
        if mon is False:
            v2 = np.array(t2mdat[n][det])
        if xsize is None:
            t3mdat.append(v2)
        else:
            try:
                t3mdat.append(interpolate.interp1d(v1, v2)(xx))
            except Exception:
                raise ValueError(
                    "error at minus helicity;", "check interpolation range."
                )

    # merging same helicities:
    t4pdat = []  # all from +hel merged
    for k in range(0, len(xx)):
        t4pdat.append(
            np.sum([t3pdat[s][k] for s in range(len(t3pdat))])
            / int(len(t3pdat))
        )
    t4mdat = []  # all from -hel merged
    for k in range(0, len(xx)):
        t4mdat.append(
            np.sum([t3mdat[s][k] for s in range(len(t3mdat))])
            / int(len(t3mdat))
        )
    t5pdat = np.array(t4pdat)
    t5mdat = np.array(t4mdat)
    pdm = t5pdat / t5mdat  # plus signal divided by minus signal

    # correlation of plus and minus signal via two linear functions:
    corrpre = np.poly1d(
        np.polyfit(xx[: int(len(xx) * 0.1)], pdm[: int(len(xx) * 0.1)], 1)
    )
    corrpost = np.poly1d(
        np.polyfit(
            xx[int(len(xx) - len(xx) * 0.1):],
            pdm[int(len(xx) - len(xx) * 0.1):],
            1,
        )
    )
    x1 = xx[2]
    x2 = xx[-2]
    cfy1 = corrpre(x1)  # corr intens. at
    cfy2 = corrpost(x2)

    # correlated signal with pmerged:
    t5mdat = t5mdat * (
        cfy1 - x1 * (cfy1 - cfy2) / (x1 - x2) + xx * (cfy1 - cfy2) / (x1 - x2)
    )
    #####
    xas = (t5pdat + t5mdat) / 2

    # Subtracting backgrounds now:
    linbkg = np.poly1d(
        np.polyfit(xx[: int(len(xx) * 0.1)], xas[: int(len(xx) * 0.1)], 1)
    )(xx)
    xas = xas - linbkg
    pxas = t5pdat - linbkg
    mxas = t5mdat - linbkg

    # normalization:

    norm_opt = ["white_line", "edge_jump", "pre_edge", "None", None]

    # choose factor:

    try:
        if norm == "white_line":
            norm_factor = float(np.max(xas))
        elif norm == "edge_jump":
            last_values = int(len(xx) * 0.05)
            norm_factor = np.mean(xas[:-last_values])

        elif norm == "pre_edge":
            last_values = int(len(xx) * 0.02)
            norm_factor = np.mean(xas[last_values:])

        elif norm in ["None", None]:
            norm_factor = 1.0

    except Exception:
        raise ValueError("normalization not clear, use {}".format(norm_opt))

    xas = np.array(xas) / norm_factor
    pxas = np.array(pxas) / norm_factor
    mxas = np.array(mxas) / norm_factor
    xmcd = np.array(pxas - mxas)

    return xx, pxas, mxas, xas, xmcd


###


# HSYT functions:

# mHYST
#
# new in 0.1.10, developed for many (mHYST) loops in one file:
# added log option in 0.1.13

# added plot together in 0.2.2 and removed typo function,
# automated ene_cut

# version 0.3.0 generalized for more possible magnetic field header names


class mHYST:
    """
    --- class was significantly changed with update 0.3.0 ---

    Arguments
    ---------

        df      = pandas dataframe of multiple Hyst loops
        fld   = field column name of pandas object df
        ene     = energy column name of pandas object df
        det     = detector column name of pandas object
        mon     = monitor column -> if no monitor wanted, put 'False'
        ratio   = either divide higher/lower or lower/higher energy
        log     = neg. logarithm for tranmission experiments, default False

    Returns
    --------
    Nothing, but contains multiple functions:
        - average_loops(), average specific loops
        - plot_separated(onefigure=False), plots all loops separated,
          if onefigure=True,
          it will plot every loop in one figure

    Notes
    --------
    no notes

    """

    def __init__(
        self, df, fld, ene, det, mon, ratio="higher/lower", log=False
    ):
        import pandas as pd

        header = list(df)
        # CHECKS:
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df is not a pd.DataFrame type")

        for n in [fld, ene, det]:
            if n not in header:
                raise ValueError(
                    "{} is not a column name of the pd.DataFrame".format(n)
                )

        if mon == "False":
            mon = False

        if mon is not False:
            if mon not in header:
                raise ValueError(
                    "{} is not a column name of the pd.DataFrame".format(mon)
                )

        # CHECKS end

        t1 = df
        self.df = df
        self.fld = fld

        # find energy for cutoff:
        ene_cut = np.around(
            (np.max(df[ene]) - np.min(df[ene])) / 2 + np.min(df[ene]), 2
        )

        try:
            # t2 at higher energy -> usually l2 edge, l3 at lower energies:
            # normalizing signal to clock
            t2, t3 = t1[t1[ene] >= ene_cut], t1[t1[ene] <= ene_cut]
            t2 = t2.reset_index()
            t3 = t3.reset_index()
        except Exception:
            raise ValueError("energy cut off failed")

        try:
            if mon is False:
                print("no monitor selected")
                t2["normlzd"] = t2[det]
                t3["normlzd"] = t3[det]
            elif log is True:
                t2["normlzd"] = -np.log(t2[det] / t2[mon])
                t3["normlzd"] = -np.log(t3[det] / t3[mon])
            else:
                t2["normlzd"] = t2[det] / t2[mon]
                t3["normlzd"] = t3[det] / t3[mon]
        except Exception:
            raise ValueError("normalization failed")

        # now use t2 to continue algortithm
        try:
            if ratio == "higher/lower":
                t2["divided"] = t2["normlzd"] / t3["normlzd"]
            elif ratio == "lower/higher":
                t2["divided"] = t3["normlzd"] / t2["normlzd"]
        except Exception:
            raise ValueError("please choose a valid ratio")

        try:
            len_t2 = len(t2)
            self.t2 = t2
            self.slope_ct = self.slope_count(t2, fld).astype(int)
            self.len_per_loop = (len_t2 / self.slope_ct).astype(int)

        except Exception:
            raise ValueError(
                "Could not build Hysteresis loop. Different number of scans? "
            )

    def slope_count(self, df, fld):
        """
        function returns the number of slopes in HYST loops
        """
        df, fld = df, fld
        magfield = np.array(df[fld])
        mag_sign = np.sign(magfield)
        signchange = ((np.roll(mag_sign, 1) - mag_sign) != 0).astype(int)
        return (np.sum(signchange) / 4).astype(int)

    def plot_seperated(self):
        raise SyntaxError("use plot_separated()")

    def plot_separated(self, onefigure=False):
        if onefigure is False:
            for n in range(self.slope_ct):
                plt.figure()
                plt.title("loop {}".format(n))
                plt.plot(
                    self.t2[self.fld][
                        n * self.len_per_loop:(n + 1) * self.len_per_loop
                    ],
                    self.t2["divided"][
                        n * self.len_per_loop:(n + 1) * self.len_per_loop
                    ],
                )
                plt.xlabel("magnetic field [arb. units]")
                plt.ylabel("absorption [arb. units]")
                plt.show()
        elif onefigure is True:
            plt.figure()
            for n in range(self.slope_ct):
                plt.plot(
                    self.t2[self.fld][
                        n * self.len_per_loop:(n + 1) * self.len_per_loop
                    ],
                    self.t2["divided"][
                        n * self.len_per_loop:(n + 1) * self.len_per_loop
                    ],
                    label="{}".format(n),
                )
            plt.xlabel("magnetic field [arb. units]")
            plt.ylabel("absorption [arb. units]")
            plt.legend(title='loop nr.')
            plt.show()
        else:
            raise ValueError("please use onefigure=True/False")

    def average_loops(self, av_list, return_data=False):

        """
        takes number of loops, check before with plot_seperated()
        set return_data to True to get data as tuple(field, signal)
        non plot mode otherwise returns plot of average loops
        """
        self.av_list = av_list
        toaverage = []
        for n in av_list:
            toaverage.append(
                self.t2["divided"][
                    n * self.len_per_loop:(n + 1) * self.len_per_loop
                ]
            )
        averaged = np.mean(toaverage, axis=0)
        self.std = np.std(toaverage, axis=0)
        field = self.t2[self.fld][: self.len_per_loop]

        if return_data is False:
            plt.figure()
            plt.title(f"average loops {*av_list,}")
            plt.plot(field, averaged)
            plt.show()

        elif return_data is True:
            return (np.array(field), np.array(averaged))
        else:
            raise ValueError("return_data mode not clear. use False or True")


def HYST3en(df, fld, ene, det, mon, Epre, Eedg1, Eedg2):
    """
    HYST : edge/pre_edge
    VEKMAG compatible
    df     = takes pandas data frame
    fld    = column name of field values
    ene    = column name of energy values
    det    = column name of detector values
    mon    = column name of monitor
    Epre    = energy of pre edge point
    Eedg1   = energy of edge point 1
    Eedg2   = energy of edge point 2
    """
    df = df
    hyst1 = []
    hyst2 = []
    field1 = []
    for n in np.arange(-6.8, 6.8, 0.1):
        n = np.around(n, 1)
        t1 = df[np.around(df[ene], 1) == Eedg1]
        t1 = t1[np.around(t1[fld], 1) == n]
        t11 = np.mean(t1[det] / t1[mon])
        t2 = df[np.around(df[ene], 1) == Eedg2]
        t2 = t2[np.around(t2[fld], 1) == n]
        t12 = np.mean(t2[det] / t2[mon])
        t1 = df[np.around(df[ene], 1) == Epre]
        t1 = t1[np.around(t1[fld], 1) == n]
        t22 = np.mean(t1[det] / t1[mon])
        hyst1.append(t11 / t22)
        hyst2.append(t12 / t22)
        field1.append(n)
    ############
    return field1, hyst1, hyst2


# sumrules functions
def sumrules_function(xx, xas00, xmcd, px, nh):
    """
    xx = global energy scale
    xas00 = xas - stepfunction
    xmcd = xmcd
    px = point in energy for p value
    nh = number of electron holes"""
    r = np.max(np.cumsum(xas00))
    q = np.max(np.cumsum(xmcd))
    p = np.cumsum(xmcd)[list(np.around(xx, 2)).index(px)]
    # print(np.max(np.cumsum(np.linspace(0,1,steps)))/steps)
    lz = nh * 2 * q / (3 * r)
    sz = nh * (3 * p - 2 * q) / (2 * r)
    return sz, lz


# old sum rule function, please do not use, thanks
def LDS(xx, xmcd, px):
    """
    xx = global energy scale
    xmcd = xmcd
    px = point in energy for p value"""
    q = np.absolute(np.cumsum(xmcd))[-1]
    p = np.absolute(np.cumsum(xmcd)[list(np.around(xx, 2)).index(px)])
    # print(q,p)
    lds = (4 / 3) * q / (3 * p - 2 * q)
    return lds


def orbital_to_spin_ratio(
    xmcd=None, xp=None, xq=None, group=None, orbital="3d"
):

    """
    function for orbital to spin ratio from xmcd without <Tz> term

    Arguments
    ---------

    xmcd    = xmcd spectra as np.array if group is none
    xp,xq   = index of XMCD for dividing (xp) and end of XMCD (xq)
              default values xp = len/2 and xq = last value
    group   = group of a sample
    orbital = probed orbital 3d or 4f

    Returns
    --------

    orbital to spin ratio attached to group if group was provided

    if no group was provided prints orbital to spin ratio

    Notes
    --------

    no notes
    """

    if xmcd is not None:
        xmcd = xmcd

    elif group is not None:
        xmcd = group.xmcd
    else:
        raise ValueError("insert XMCD as group or numpy array")

    if xp is None:
        xp = int(len(xmcd) / 2)

    A = np.cumsum(xmcd[:xp])[-1]
    B = np.cumsum(xmcd[xp:xq])[-1]

    def lds_self_3d(A, B):
        return (2 / 3) * ((A + B) / (A - 2 * B))

    def lds_self_4f(A, B):
        return (A + B) / ((A - (3 / 2) * B))

    if orbital == "3d":
        if group is None:
            return lds_self_3d(A, B)

        if group is not None:
            group.lds = lds_self_3d(A, B)

    elif orbital == "4f":
        if group is None:
            return lds_self_4f(A, B)

        if group is not None:
            group.lds = lds_self_4f(A, B)

    else:
        raise ValueError("choose a valid probed orbital: 3d or 4f")


def Lz(
    xmcd, xas, c=1, l=2, nh=1, tz=None, last_number_xas=1, last_number_xmcd=1
) -> float:
    """
    sum rule taken from:
        Carra, P., Thole, B. T., Altarelli, M., & Wang, X. (1993).
        X-ray circular dichroism and local magnetic fields.
        Phys Rev Lett, 70(5), 694-697.
        doi:10.1103/PhysRevLett.70.694
    xmcd = numpy array of XMCD as in mu_p - mu_m
    xas  = numpy array of step function corrected XAS as in (mu_p + mu_m)/2
    c    = s:0, p:1, d:2, f:3, initial orbital
    l    = s:0, p:1, d:2, f:3, final orbital
    nh   = number of holes
    tz   = can be provided, but will not be used for calculating orbital
           moment
    """
    tz = tz

    factor = (1 / 2) * ((l * (l + 1)) - (c * (c + 1)) + 2) / ((l * (l + 1)))
    # xas is (mu_p + mu_m)/2
    lz = (nh / factor) * (
        np.cumsum(xmcd)[-last_number_xmcd]
        / (3 * np.cumsum(xas)[-last_number_xas])
    )
    return float(lz)


def Sz(
    xmcd,
    xas,
    c=1,
    l=2,
    nh=1,
    tz=0,
    edge_div=None,
    last_number_xas=1,
    last_number_xmcd=1,
) -> float:
    """
    sum rule taken from:
        Carra, P., Thole, B. T., Altarelli, M., & Wang, X. (1993).
        X-ray circular dichroism and local magnetic fields.
        Phys Rev Lett, 70(5), 694-697.
        doi:10.1103/PhysRevLett.70.694
    xmcd = numpy array of XMCD as in mu_p - mu_m
    xas  = numpy array of step function corrected XAS as in (mu_p + mu_m)/2
    c    = s:0, p:1, d:2, f:3, initial orbital
    l    = s:0, p:1, d:2, f:3, final orbital
    nh   = number of holes
    tz   = magnetic dipole term
    edge_div = if None, middle of spectrum, relative shift to calculated
               middle of spectrum based on index number
    """
    if edge_div is None:
        x_div = int(len(xmcd) / 2)
    else:
        x_div = int(len(xmcd) / 2) + edge_div

    tz_part = (
        (
            l * (l + 1) * (l * (l + 1) + 2 * c * (c + 1) + 4)
            - ((3 * (c - 1) ** 2) * (c + 2) ** 2)
        )
        / (6 * l * c * (l + 1) * nh)
    ) * tz

    spec_part = (
        np.cumsum(xmcd[:x_div])[-1]
        - (((c + 1) / c) * np.cumsum(xmcd[x_div:])[-last_number_xmcd])
    ) / (3 * np.cumsum(xas)[-last_number_xas])

    sz = (spec_part - tz_part) * (
        (3 * c * nh) / ((l * (l + 1) - 2 - c * (c + 1)))
    )
    return float(sz)



# module imports:
import pandas as pd
import numpy as np
import os
import pickle
import warnings

# readin functions

# for beamline 6.3.1 at ALS, Berkeley


def count_lines(file):
    header = []
    f = open(file)
    ct = 0
    for n in range(50):
        line = str(f.readline())
        if line.startswith("Time of Day"):
            break
        ct += 1
        header.append(line)
    t1 = pd.read_csv(file, skiprows=ct - 1, sep="\t", engine="python")
    f.close()
    return t1, header


def SS_indexing():
    datapath = "Y:\\BCS Setup Data\\"
    SigScans_indexed = {}
    for root, dirs, files in os.walk(datapath):
        for name in files:
            if name.startswith("SigScan"):
                strng = name.replace(".txt", "")
                strng = strng.replace("SigScan", "")
                SigScans_indexed.update({strng: os.path.join(root, name)})
    return SigScans_indexed


def TS_indexing():
    datapath = "Y:\\BCS Setup Data\\"
    TrajScans_indexed = {}
    for root, dirs, files in os.walk(datapath):
        for name in files:
            if name.startswith("TrajScan"):
                strng = name.replace(".txt", "")
                strng = strng.replace("TrajScan", "")
                TrajScans_indexed.update({strng: os.path.join(root, name)})
    return TrajScans_indexed


class SigScan:
    # "bl_comp" working at beam line computer with access to all scans
    path = "bl_comp"
    # import using the beam line computer

    def __init__(self, string: str):
        if self.path == "bl_comp":
            try:
                indexed = pickle.load(open("_SS_index.p", "rb"))
                file_to_open = indexed[string]

            except Exception:
                raise ValueError(
                    "could not find Single Scan "
                    + string
                    + ". Typo? Indexing server running?"
                )
            try:
                ct_lines = count_lines(file_to_open)
                self.df = ct_lines[0]
                self.header = ct_lines[1]
            except Exception:
                raise ValueError("error while reading or finding file")
            try:
                self.scantype = guess_scan(self.df)
            except Exception:
                self.scantype = "could not indentify scan type"

        ###
        # import on local machine
        if self.path != "bl_comp":
            if os.path.isdir(self.path) is not True:
                raise ValueError("could not find local path")

            for root, dirs, files in os.walk(self.path):
                for name in files:
                    if name.startswith("SigScan"):
                        if string in name:
                            ct_lines = count_lines(os.path.join(root, name))
                            self.df = ct_lines[0]
                            self.header = ct_lines[1]
                            try:
                                self.scantype = guess_scan(self.df)
                            except Exception:
                                self.scantype = "could not indentify scan type"


class TrajScan:
    # "bl_comp" working at beam line computer with access to all scans
    path = "bl_comp"
    # import using the beam line computer

    def __init__(self, string: str):
        if self.path == "bl_comp":
            try:
                indexed = pickle.load(open("_TS_index.p", "rb"))
                file_to_open = indexed[string]

            except Exception:
                raise ValueError(
                    "could not find Trajectory Scan "
                    + string
                    + ". Typo? Indexing server running?"
                )
            try:
                ct_lines = count_lines(file_to_open)
                self.df = ct_lines[0]
                self.header = ct_lines[1]
            except Exception:
                raise ValueError("error while reading or finding file")
            try:
                self.scantype = guess_scan(self.df)
            except Exception:
                self.scantype = "could not indentify scan type"

        ###
        # import on local machine
        if self.path != "bl_comp":
            if os.path.isdir(self.path) is not True:
                raise ValueError("could not find local path")

            for root, dirs, files in os.walk(self.path):
                for name in files:
                    if name.startswith("TrajScan"):
                        if string in name:
                            ct_lines = count_lines(os.path.join(root, name))
                            self.df = ct_lines[0]
                            self.header = ct_lines[1]
                            try:
                                self.scantype = guess_scan(self.df)
                            except Exception:
                                self.scantype = "could not indentify scan type"


# returns type of scan
# this can be written more elegant
def guess_scan(df: pd.DataFrame, check: bool = False) -> str:

    """
    guesses scan from pandas input base on number of unique scan values

    returns None if no column was written
    """

    len_magfield = len(np.unique(np.around(df["Magnet Field"], 3)))
    try:
        len_energy = len(np.unique(np.around(df["Energy"], 1)))
    except Exception:
        len_energy = 1
    len_x = len(np.unique(np.around(df["X"], 2)))
    len_y = len(np.unique(np.around(df["Y"], 2)))
    len_z = len(np.unique(np.around(df["Z"], 2)))
    len_theta = len(np.unique(np.around(df["Theta"], 2)))

    if check is True:
        print("magnetic field points: {}".format(len_magfield))
        print("energy points: {}".format(len_energy))
        print("xmotor points: {}".format(len_x))
        print("ymotor points: {}".format(len_y))
        print("zmotor points: {}".format(len_z))
        print("theta motor points: {}".format(len_theta))

    parameter_list = [len_magfield, len_energy, len_x, len_y, len_z, len_theta]

    if 0 in parameter_list:
        return None

    most_changed = parameter_list.index(max(parameter_list))

    answers = ["Magnetic Field", "Energy", "X", "Y", "Z", "Theta"]

    return answers[most_changed]


# checks if imported pd.DataFrame object is a Hysteresis loop measurement
def ishyst(df, check=False):
    len_magfield = len(np.unique(np.around(df["Magnet Field"], 3)))
    len_energy = len(np.unique(np.around(df["Energy"], 1)))

    if check is True:
        print(
            "magnetic field points: {}, energy points: {}".format(
                len_magfield, len_energy
            )
        )
    try:
        if len_magfield > len_energy:
            return True
        elif len_magfield < len_energy:
            return False
    except Exception:
        raise ValueError("unknown Scan")


#######
# Make scanfiles


def print_mesh(mesh_fields, i=0, line="", outfile=None):
    if i < len(mesh_fields):
        for v in mesh_fields[i]["Values"]:
            if v == "file":
                outfile.write("file")
                outfile.write("\r\n")
            else:
                # Use join next time
                if i == 0:
                    tab = ""
                else:
                    tab = "\t"
                print_mesh(mesh_fields, i + 1, line + tab + str(v), outfile)
    else:
        outfile.write(line)
        outfile.write("\r\n")


def XMCD_scanpair(
    field: float, energy: list, step=0.1, velocity=1.0, pairs: int = 4
) -> list:
    """

    Produces a Trajectory Scan File for Beamlines 6.3.1 and 4.0.2 at the ALS

    Arguments
    ---------

        field    = field for XMCD in Tesla
        energy   = takes energy range as list, e.g. [765,815]
        step     = stepsize for flying Energy scans, default 0.1
        velocity = velocity for flying Energy scans, default 1.0
        pairs    = number of XMCD pairs, has to be an integer, default 4

    Returns
    --------

    list of strings, input for mesh_fields

    Notes
    --------

    - new file will be created after two opposite fields
    - fields will be like + - - + if more than 1 pair

    """

    if abs(field) > 1.9:
        raise Warning(
            "Most probably not possible for this setup, check the magnetic field value again."
        )

    if not isinstance(pairs, int):
        pairs = int(pairs)
        raise Warning(
            "Number of pairs must be an integer value, e.g. 4., continued with {} pairs".format(
                pairs
            )
        )

    if len(energy) != 2:
        raise ValueError(
            "Number of energy range has to be 2, takes energy range as list, e.g. [765,815]"
        )

    if energy[0] > energy[1]:
        warnings.warn("Scan is running backwards in energy.")
    elif energy[0] == energy[1]:
        warnings.warn("Same point for start and stop energy.")

    sl = list()
    for i in range(pairs):
        sl.append(
            "{}{}\tflying({},{},{},{})".format(
                ["+", "-"][i % 2], field, energy[0], energy[1], step, velocity
            )
        )
        sl.append(
            "{}{}\tflying({},{},{},{})".format(
                ["-", "+"][i % 2], field, energy[0], energy[1], step, velocity
            )
        )
        sl.append("file")
    return sl


def HYST_scanpair():
    """
    DUMMY for HYST scans
    """
    pass


def make_scanfile(mesh_fields, outfilename):
    with open(outfilename, "w", newline="") as outfile:
        if "flying" in str(mesh_fields):
            if "Energy" in str(mesh_fields):
                # Energy is set explicitly, so assume Hysteresis with flying field
                # dummy values to introduce the right flying scan
                outfile.write(
                    "flying Magnet Field(0.41, -0.41, 0.01,0.1)\t\r\n",
                )
            else:
                # Energy is not set explicitly, so assume that is what we are flying for XAS or XMCD
                # dummy values to introduce the right flying scan
                outfile.write(
                    "flying Energy(700, 800, 0.1,1)\r\n",
                )
        for h in mesh_fields:
            outfile.write(h["Header"])
        outfile.write("\r\n")
        print_mesh(mesh_fields, outfile=outfile)



#### OLD namespace -> moved to beamlines


def raise_error():
    raise ValueError("beam line 6.3.1 functions now in xaspy.beamlines.bl631")


def count_lines(file):
    return raise_error()


def SS_indexing():

    return raise_error()


def TS_indexing():

    return raise_error()


class SigScan:
    ### "bl_comp" working at beam line computer with access to all scans
    path = "bl_comp"
    # import using the beam line computer

    def __init__(self, string: str):
        return raise_error()


class TrajScan:
    ### "bl_comp" working at beam line computer with access to all scans
    path = "bl_comp"
    # import using the beam line computer

    def __init__(self, string: str):
        return raise_error()



import numpy as np
import pandas as pd


class get_scan_metadata:
    def __init__(self, file):
        if "TrajScan" in file:
            t1 = pd.read_csv(file, skiprows=9, sep="\t", engine="python")
            self.df = t1
            # print the values:
            self.energy(t1)
            self.field(t1)
            self.temperature(t1)
            self.sample_position(t1)

    def energy(self, df):
        """
        get properties of energy steps
        """
        df = df
        arr = np.array(df["Energy"])

        # check if HYST oder XMCD, XMCD field is const
        # HYST field changing check for first 10 datapoints
        step = []
        for n in range(10):
            step.append(np.around(arr[n + 1], 1) - np.around(arr[n], 1))
        step_mean = np.around(np.mean(step), 1)
        step_std = np.around(np.std(step), 1)

        if (step_mean == 0) and (step_std == 0):
            unique_ene = np.unique(np.around(arr, 1))
            meas_type = f"HYST at points {unique_ene}eV"

        elif (step_mean != 0) and (step_std == 0):
            meas_type = f"energy scan with step size {step_mean}eV"

        else:
            meas_type = "unknown measurement type"

        arr_min, arr_max = np.around(np.min(arr), 1), np.around(np.max(arr), 1)
        print(f"{arr_min}eV and {arr_max}eV {meas_type}")

    def field(self, df):
        """
        get properties of magnetic field
        """
        df = df
        arr = np.array(df["Magnet Field"])

        # check if HYST oder XMCD, XMCD field is const
        # HYST field changing check for first 10 datapoints
        step = []
        for n in range(10):
            step.append(np.around(arr[n + 1], 3) - np.around(arr[n], 3))
        step_mean = np.around(np.mean(step), 3)
        step_std = np.around(np.std(step), 3)

        if (step_mean == 0) and (step_std == 0):
            meas_type = "static field(s)"

        elif (step_mean != 0) and (step_std == 0):
            meas_type = f"HYST with step size {step_mean}T"
        else:
            meas_type = "unknown measurement type"

        arr_min, arr_max = np.min(arr), np.max(arr)
        print(f"min: {arr_min}T | max: {arr_max}T {meas_type}")

    def temperature(self, df):
        """
        get properties of temperature sensor Temp B
        """
        df = df
        arr = np.array(df["Temp B"])

        temp = np.around(np.mean(arr), 1)
        print(f"measurement temperature {temp}K")

    def sample_position(self, df):
        """
        get x,y,z coordinates for sample determination, and angle
        """
        df = df

        list_pos = []
        for n in ["X", "Y", "Z", "Theta"]:
            arr_mean = np.mean(np.array(df[n]))
            list_pos.append(np.around(arr_mean, 1))
        print("sample position at [X,Y,Z,Theta]")
        print(list_pos)


def toggle(**kws):
    """
    removed in version 0.2.1:
        toggles cell in Jupyter notebook;
        makes button with label 'a'
    """
    return print("this has been removed in version 0.2.1")


def showspeccom(a):
    """
    a = SPEC file
    returns list with commands/comments #C and #S
    """
    clst = []
    with open(a, "r") as f:
        for line in f:
            if "#C" in line:
                clst.append(line.replace("\n", ""))
            if "#S" in line:
                clst.append(line.replace("\n", ""))
    return clst


def center(a: float, b: float) -> int:
    """
    get position a and b, returns center between them

    Arguments
    ---------
        a   = number, can be float, should be index-like
        b   = number, can be float, should be index-like

    Returns
    --------
        c   = center between a and b as int

    Notes
    --------
    no notes
    """
    c = int((b - a)/2) + a
    return c



import pandas as pd
import json
import numpy as np
import os
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("echo")
args = parser.parse_args()

# Opening JSON file from input:
filename = args.echo

# Read in the file
with open(filename + ".txt", "r") as file:
    filedata = file.read()

# Replace the target string
filedata = filedata.replace("'", '"')

# Write the file out again
with open(filename + ".txt", "w") as file:
    file.write(filedata)


f = open(filename + ".txt")

data = json.load(f)
f.close()


cutoff = int(len(np.array(data["MCA_3"])) - len(np.array(data["I0"])))

smaller_length = len(np.array(data["I0"]))


# also cut out the last 5 points turned out to be a good approx, for different reasons and convenience
for n in list(data):
    if n == "timePP":
        data[n] = np.ones(smaller_length - 5) * data[n]

    elif len(data[n]) != smaller_length:
        data[n] = np.array(data[n])[cutoff:-5]
    else:
        data[n] = np.array(data[n])[:-5]

df = pd.DataFrame.from_dict(data)
# print(len(list(data)))

# check if this is only header by filesize:
if os.stat(filename + ".xdi").st_size <= 20 * 1024:

    with open(filename + ".xdi", "a") as f:
        dfAsString = df.to_string(header=False, index=False)
        f.write(dfAsString)
    f.close()



# SPEC file splitter
#
# author gnzng
#
# small python program to split a larger SPEC file into individual files
#
# new empty folder for individual scans has to be created in the same directory. eg. scans_individual
#
# splitted files are .dat files
# contain first line scan number and command
# second line name of detectors/header


import re


# input parameters:
file = input("Enter Filename:")
output_folder = input('write scans to folder (folder has to already exist):')
# end input parameters

# split to remove extension from filename
filename = file.split('.')[0]


def split(file):
    clean = []
    clst = []
    with open(file, "r") as f:
        regexp = re.compile(r"^-?\d+\.?\d*")
        for line in f:
            if "#C" in line:
                clst.append(line.replace("\n", ""))
            if "#S" in line:
                clean.append(line.replace("\n", ""))
                nr = line.split(" ")[1]
            if "#L" in line:
                clean.append(line[3:].replace("\n", ""))
            if regexp.search(line):
                clean.append(line.replace("\n", ""))
            if not line.strip():
                clean.append("splitsignal")
    for n in range(1, 10000):
        if clean[0] == "splitsignal":
            clean.pop(0)
        else:
            break
    nr = 0
    for i in clean:
        if "#S" in i:
            nr = i.split(" ")[1]
            f = open(
                output_folder + "/" + filename + "_{0:03}.dat".format(int(nr)),
                "w+",
            )
        f = open(
            output_folder + "/" + filename + "_{0:03}.dat".format(int(nr)), "a"
        )
        if i != "splitsignal":
            f.write(i + "\n")
        if n != 0:
            if i == "splitsignal":
                f.close()


split(file)

print("excuted")



# import:
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


class PiMoss:
    """
    ---under construction---
    class to create Mossbauer measurement class from pi software package:
    https://www.uni-due.de/~hm236ap/hoersten/home.html
    takes path to .dat file. do not use .dat extension in name
    """

    def __init__(self, path):
        def impspec(a):
            """
            function to import the plot data as a pandas DataFrame
            """
            return pd.read_csv(
                a,
                sep="\t",
                engine="python",
                skiprows=1,
                na_values="-",
                header=None,
            )

        try:
            self.df = impspec(path + ".dat")
            self.v = np.array(self.df[0])  # returns velocity as numpy array
            self.m = np.array(self.df[1])  # returns measurement as numpy array
            self.err = np.array(self.df[2])  # returns velocity as numpy array
            self.lsfit = []  # list of np.arrays for the fit models
            for n in range(len(self.df.columns) - 3):
                self.lsfit.append(np.array(self.df[n + 3]))
        except:
            raise IOError("File not found. Make sure it is a .dat-file")

    def plot(self, kind="all"):
        """
        fast plotting functions functions:
        input string:
        'all'        errorbar measurement + all fit models
        'err'        errorbar measurement
        'noerr'      measurement no errorbar
        'allnoerr'   all fit models + measurement no errorbar
        """

        plotlist = ["all", "err", "noerr", "allnoerr"]

        if kind not in plotlist:
            raise ValueError(
                "Plot option not in list. Please choose from {}".format(
                    plotlist
                )
            )
        if kind == "all":
            plt.errorbar(self.v, self.m, self.err)
            for n in self.lsfit:
                plt.plot(self.v, n)
        if kind == "err":
            plt.errorbar(self.v, self.m, self.err)
        if kind == "noerr":
            plt.plot(self.v, self.m)
        if kind == "allnoerr":
            plt.plot(self.v, self.m)
            for n in self.lsfit:
                plt.plot(self.v, n)
        plt.xlabel("velocity [mm/s]"), plt.ylabel("absorption [arb. units]")



from ._version import __name__, __date__, __version__, __authors__


class info:
    """
    simple class for printing package info all at once
    """

    def __init__(self):
        print("name:", "\t", __name__)
        print("updated:", "\t", __date__)
        print("version:", "\t", __version__)
        print("authors:", "\t", __authors__)



from xas import group

import unittest

import numpy as np

xmcd_141_ = group("test", lds=0.7)
xmcd_141_.energy = np.linspace(0, 1, 5000)
xmcd_141_.energy_interp = np.linspace(0, 1, 10000)
xmcd_141_.plus_xas = np.array([1, 2, 3, 4, 5])
xmcd_141_.minus_xas = np.array([1, 2, 3, 4, 5])

assert xmcd_141_, "group should be existent"
assert xmcd_141_.lds == 0.7, "should be 0.7"
assert xmcd_141_.__name__ == "test", "name should be test"


class Test(unittest.TestCase):
    def test_1(self):
        self.assertEqual(xmcd_141_.lds, 0.7, "should be 0.7")

    def test_2(self):
        self.assertEqual(xmcd_141_.__name__, "test", "name should be test")


if __name__ == "__main__":
    unittest.main()



import unittest

import pickle
import numpy as np
import pandas as pd
import pytest
import os

# imports from xaspy:

from polarized import (
    orbital_to_spin_ratio,
    mHYST,
)
from xas import group

# HYST tests:

path_to_current_file = os.path.realpath(__file__)
current_directory = os.path.split(path_to_current_file)[0]
path_to_file = os.path.join(
    current_directory, "../test_files/pd_dataframe_hyst_example1.pickle"
)
df = pickle.load(open(path_to_file, "rb"))


def test_import_df():
    # nr: 'TrajScan31914-001_0001.txt'
    assert isinstance(df, pd.DataFrame)


hystclass = mHYST(
    df, "Magnet Field", "Energy", "LY", "Clock", ratio="lower/higher", log=True
)


def test_hyst_from_df():
    assert hystclass
    assert isinstance(
        hystclass.average_loops([n for n in range(1, 16)], return_data=True),
        tuple,
    )


def test_raise_value_error_bc_wrong_column_name():
    with pytest.raises(ValueError):
        mHYST(
            df,
            "_non_existent_column",
            "Energy",
            "LY",
            "Clock",
            ratio="lower/higher",
            log=True,
        ).average_loops([n for n in range(1, 16)], return_data=True)


# XMCD tests


testgroup = group()

path_to_current_file = os.path.realpath(__file__)
current_directory = os.path.split(path_to_current_file)[0]
path_to_file = os.path.join(current_directory, "../test_files/xmcd_co.pickle")
testgroup.xmcd = pickle.load(open(path_to_file, "rb"))

testxmcd = testgroup.xmcd


class Test(unittest.TestCase):
    # in unit test orbital to spin ratio must be imported

    def test_errors(self):
        #
        # test if errors are right
        #
        with self.assertRaises(ValueError):
            orbital_to_spin_ratio(xmcd=testxmcd, orbital="11s")
        with self.assertRaises(ValueError):
            orbital_to_spin_ratio(orbital="11s")

    def test_lds(self):
        #
        # test if orbital_to_spin values yields real values
        #

        # test for xmcd testset:
        self.assertEqual(
            np.around(0.10483034490920523, 1),
            np.around(
                orbital_to_spin_ratio(
                    xmcd=testxmcd, group=None, xp=2750, orbital="3d"
                ),
                1,
            ),
        )

        # test for group:
        orbital_to_spin_ratio(
            xmcd=None, group=testgroup, xp=2750, orbital="3d"
        )
        self.assertEqual(
            np.around(0.10483034490920523, 1), np.around(testgroup.lds, 1)
        )


if __name__ == "__main__":
    unittest.main()



# imports:
import pandas as pd

#####

####readin functions for different theory/simulation programs:


def multix(a):
    """
    import multiX output file
    returns panda file with energy, imaginary part that simulates the xas, and the real part
    """
    return pd.read(a, header=["energy", "img", "r"], delim_whitespace=True)


def crispy(a):
    return pd.read_csv(a, delim_whitespace=True, comment="#", header=3)



import pickle

###########################
###########################
####DESPIKE Functions######
###########################
###########################
def showspikes(file):
    """
    shows dictionary of spike values saved in file
    file = path to spike file, without -.spike-
    """
    with open(file + ".spike", "rb") as f:
        b = pickle.load(f)
    return print(b)


def removespikes(a, file):
    """
    removes the spike values for scan a
    a = scan number, where all scans are removed
    file = path to spike file, without -.spike-
    """
    with open(file + ".spike", "rb") as f:
        b = pickle.load(f)
    try:
        del b[a]
        with open(file + ".spike", "wb") as f:
            pickle.dump(b, f, protocol=pickle.HIGHEST_PROTOCOL)
    except:
        print("{} not in {}".format(a, file))


def addspike(a, c, file):
    """
    adds one  spike value c for scan a
    a = scan number, where all scans are removed
    file = path to spike file, without -.spike-
    """
    try:
        with open(file + ".spike", "rb") as f:
            b = pickle.load(f)
    except:
        with open(file + ".spike", "wb") as f:
            pickle.dump({}, f, protocol=pickle.HIGHEST_PROTOCOL)
        with open(file + ".spike", "rb") as f:
            b = pickle.load(f)
    try:
        if a in b:
            if c not in b[a]:
                b[a].append(c)
        else:
            b[a] = [c]
        with open(file + ".spike", "wb") as f:
            pickle.dump(b, f, protocol=pickle.HIGHEST_PROTOCOL)
    except:
        print("oops, something went wrong")



import os
from distutils.core import setup
__authors__  = 'Damian Guenzing \nAlpha T. N\'Diaye'


_version__ = None
with open(os.path.join('xaspy', '_version.py'), 'r') as version_file:
    lines = version_file.readlines()
    for line in lines:
        line = line[:-1]
        if line.startswith('__version__'):
            key, vers = [w.strip() for w in line.split('=')]
            __version__ = vers.replace("'",  "").replace('"',  "").strip()


setup(
  name = 'xaspy',         # How you named your package folder (MyLib)
  packages = ['xaspy',
              'xaspy.utils',
              'xaspy.xas',
              'xaspy.readin',
              'xaspy.beamlines',], 
  version = __version__,      
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'package for analysis of experimental xray absorption spectroscopy data',   # Give a short description about your library
  author = __authors__,                     # Type in your name
  author_email = 'gnzng@protonmail.ch',      # Type in your E-Mail
  url = 'https://github.com/gnzng/xaspy',   # Provide either the link to your github or to your website
  keywords = ['xray absorption spectroscopy', 'xmcd', 'synchrotron'],   
  install_requires=[ 
          'numpy',
          'pandas',
          'scipy',
          'matplotlib',
          'python-telegram-bot',
          'xraydb'
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.7',
  ],
)


# imports:
import numpy as np
from scipy import interpolate
import matplotlib.pyplot as plt
import pandas as pd
import pickle

#####

####readin functions from different beamlines:
##VEKMAG### BESSY2
### define get header function from large, unsplit and file
def getheader(b, file):  # its written that b is scan number!
    r = open(file, "r")
    mylist = r.read().splitlines()
    matching = [p for p in mylist if "#L" in p]
    scannr = [p for p in mylist if "#S" in p]
    r.close()
    rscan = []
    for n in scannr:
        v = n.split()[1]
        rscan.append(int(v))
    header = []
    for n in range(len(rscan)):
        t1 = [p for p in matching[n].split() if not "#L" in p]
        for n in range(len(t1)):
            t1[n] = "{}_".format(n + 1) + t1[n]
        header.append(t1)
    c = rscan.index(b)
    return header[c]


def getcommand(b, file):
    r = open(file, "r")
    mylist = r.read().splitlines()
    scannr = [p for p in mylist if "#S" in p]
    r.close()
    rscan = []
    comma = []
    for n in scannr:
        v = n.split()[1]
        w = " ".join(n.split()[2:])
        rscan.append(int(v))
        comma.append(w)
    c = rscan.index(b)
    return comma[c]


# function for read in a is number of scan
def readin(a, file, raw=False):
    a = a
    dff = pd.read_csv(
        file + "_{}.dat".format(a),
        delim_whitespace=True,
        header=0,
        names=getheader(a, file),
        comment="#",
    )
    if raw == False:
        try:
            with open(file + ".spike", "rb") as f:
                b = pickle.load(f)
                if a in b:
                    todrop = b[a]
                    dff = dff.drop(todrop)
        except:
            pass
    else:
        pass
    return dff



class group:
    """
    Group: a container for variables, modules, and subgroups.
    heavy influenced by xraylarch https://github.com/xraypy/xraylarch
    """

    __private = ("__private", "_repr_html_")

    def __init__(self, name=None, **kws):
        if name is None:
            name = hex(id(self))
        self.__name__ = name
        for key, val in kws.items():
            setattr(self, key, val)

    def __id__(self):
        return id(self)

    def __dir__(self):
        """
        return list of member names, attached variable spectras and so on
        """
        cls_members = []

        dict_keys = [key for key in self.__dict__]

        return [
            key
            for key in cls_members + dict_keys
            if (
                not key.startswith("_Group_")
                and not (key.startswith("__") and key.endswith("__"))
                and key not in self.__private
            )
        ]

    def _repr_html_(self):
        """
        HTML representation for Jupyter notebook
        """
        html = [f"group name: {self.__name__} | id:{self.__id__()}"]
        html.append("<table>")
        html.append(
            "<tr><td><b>attribute</b></td><td><b>type</b></td><td><b>value</b></td></tr>"
        )
        attrs = self.__dir__()
        atypes = [type(getattr(self, attr)).__name__ for attr in attrs]
        atvalues = [getattr(self, attr) for attr in attrs]

        html.append(
            "".join(
                [
                    f"<tr><td>{attr}</td><td><i>{atp}</i></td><td><i>{atv}</i></td></tr>"
                    for attr, atp, atv in zip(attrs, atypes, atvalues)
                ]
            )
        )
        html.append("</table>")
        return "".join(html)



# imports:
import numpy as np
from scipy import interpolate
import matplotlib.pyplot as plt
import pandas
import pickle

#####


def linprebkg(xx, xas, x1, x2):
    """
    xx = global energy scale
    xas = xas
    use from x1 to x2 for cutting background region"""
    return np.poly1d(
        np.polyfit(
            xx[
                list(np.around(xx, 2))
                .index(x1) : list(np.around(xx, 2))
                .index(x2)
            ],
            xas[
                list(np.around(xx, 2))
                .index(x1) : list(np.around(xx, 2))
                .index(x2)
            ],
            1,
        )
    )(xx)


def norm01(
    xx, xas, energy1, a1
):  # a1 nimmt list object norm to one at energy1
    faktor = 1 / xas[list(np.around(xx, 2)).index(energy1)]
    return a1 * faktor


def smooth(y, box_pts):
    """
    smoothing function
    y =  function to smooth
    box_pts = points to average
    """
    box = np.ones(box_pts) / box_pts
    y_smooth = np.convolve(y, box, mode="same")
    #
    y_smooth_mean1 = np.mean(y_smooth[box_pts : box_pts + 10])
    y_smooth[: int(box_pts)] = y_smooth_mean1
    #
    y_smooth_mean2 = np.mean(y_smooth[-int(box_pts + 10) :])
    y_smooth[-int(box_pts) :] = y_smooth_mean2
    #
    return y_smooth


# step function for branching ratio
def step(a, tp, energy, slope=None):
    """
    takes:
    a      = step hight
    tp     = turning point of step function
    energy = numpy array
    #optional:
    slope = slope factor

    returns:
    step function as a np.array
    """
    if slope == None:
        return np.array(a * (1 / (1 + np.exp(tp - energy))))

    if type(slope) == float:
        return np.array(a * (1 / (1 + np.exp(slope * (tp - energy)))))
    else:
        raise ValueError("slope factor has to be float, e.g. 2.0, 2.3, etc.")


###index function
def idx(a1, b1):
    """find index of value a1 in list b1"""
    return list(np.around(b1, 2)).index(a1)



import os
from bl631 import TrajScan
from bl631 import HYST_scanpair, XMCD_scanpair


path_to_current_file = os.path.realpath(__file__)
current_directory = os.path.split(path_to_current_file)[0]
path_to_file = os.path.join(current_directory, "../test_files")


def test_imports():
    TrajScan.path = path_to_file
    TrajScan("31914-001_0001").df
    TrajScan("31914-001_0001").header
    assert TrajScan("31914-001_0001").scantype == "Magnetic Field"


def test_scn_maker():
    HYST_scanpair()
    XMCD_scanpair(1.9, [735, 800])
    pass


def test_warnings():
    #
    # test if errors are right
    assert XMCD_scanpair(1.9, [735, 734])
    assert XMCD_scanpair(1.9, [735, 735])
